# Copyright (c) 2009-2010 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

import os
import sys
import logging
import cPickle as pickle
from os.path import join, exists

from xpcom import components
from xpcom.server import WrapObject, UnwrapObject

import mozutils
from pref_serialization import PreferenceSerializer

log = logging.getLogger('koPublishingSettings')
#log.setLevel(logging.DEBUG)


class LocalSyncInfo(object):
    def __init__(self, stat):
        self.ino = stat.st_ino
        self.mode = stat.st_mode
        self.mtime = stat.st_mtime
        self.size = stat.st_size
        # Required?
        #self.uid = stat.st_uid
        #self.gid = stat.st_gid

    def __repr__(self):
        return "mode: %r, size: %r, mtime: %r, ino: %r" % (self.mode, self.size,
                                                           self.mtime, self.ino)

    def hasChanged(self, stat):
        return (self.ino != stat.st_ino or
                self.mode != stat.st_mode or
                self.mtime != stat.st_mtime or
                self.size != stat.st_size)

class RemoteSyncInfo(object):
    def __init__(self, rf_info):
        rf_info = UnwrapObject(rf_info)
        self.mode = rf_info.get_mode()
        self.mtime = rf_info.get_mtime()
        self.size = rf_info.get_size()
        # Required?
        #self.uid = rf_info.get_uid()
        #self.gid = rf_info.get_gid()

    def __repr__(self):
        return "mode: %r, size: %r, mtime: %r" % (self.mode, self.size, self.mtime)

    def hasChanged(self, rf_info):
        rf_info = UnwrapObject(rf_info)
        return (self.mode != rf_info.get_mode() or
                self.mtime != rf_info.get_mtime() or
                self.size != rf_info.get_size())

class koPublishingSettings(PreferenceSerializer):
    _com_interfaces_ = [components.interfaces.koIPublishingSettings,
                        components.interfaces.koIPreferenceSerializer]
    _reg_desc_ = "Komodo Publishing Settings"
    _reg_contractid_ = "@activestate.com/koPublishingSettings;1"
    _reg_clsid_ = "{42ead6d1-78e0-45b3-a837-49b4f1b62147}"

    # Preference map for PreferenceSerializer.
    pref_serialize_items = [
        ("id",              "string"),
        ("name",            "string"),
        ("local_uri",       "string"),
        ("remote_uri",      "string"),
        ("includes",        "string"),
        ("excludes",        "string"),
    ]

    _id = None
    _sync_data = None
    _sync_data_filepath = None

    def __init__(self, name="", local_uri="", remote_uri="", includes="",
                 excludes=""):
        PreferenceSerializer.__init__(self, self.pref_serialize_items)
        self.name = name
        self.local_uri = local_uri
        self.remote_uri = remote_uri
        # TODO: Have our own prefs for these?
        globalPrefs = components.classes["@activestate.com/koPrefService;1"]. \
                            getService(components.interfaces.koIPrefService).prefs
        if not includes:
            includes = globalPrefs.getStringPref("import_include_matches")
        if not excludes:
            excludes = globalPrefs.getStringPref("import_exclude_matches")
        self.includes = includes
        self.excludes = excludes

    @property
    def id(self):
        # Lazily generate the id as required.
        if self._id is None:
            self._id = mozutils.generateUUID()
        return self._id

    @id.setter
    def id(self, val):
        self._id = val
        # Invalidate the sync data filepath as well.
        self._sync_data_filepath = None

    @property
    def syncDataFilepath(self):
        # Lazily generate the filepath as required.
        if self._sync_data_filepath is None:
            koDirSvc = components.classes["@activestate.com/koDirs;1"].\
                        getService(components.interfaces.koIDirs)
            # All sync data is stored under the Komodo user data dir.
            publishing_data_dir = join(koDirSvc.userDataDir, "publishing")
            if not exists(publishing_data_dir):
                os.mkdir(publishing_data_dir, 0700)
            self._sync_data_filepath = join(publishing_data_dir, self.id)
        return self._sync_data_filepath

    @property
    def sync_data(self):
        # Lazily generate the sync data as required.
        if self._sync_data is None:
            if exists(self.syncDataFilepath):
                try:
                    pickle_file = file(self.syncDataFilepath, "rb")
                except Exception, ex:
                    log.warn("Could not open publishing data file %r. Error: %r",
                             self.syncDataFilepath, str(ex))
                else:
                    try:
                        self._sync_data = pickle.load(pickle_file)
                    except pickle.PickleError:
                        log.warn("Failed to load publishing data for %r. Error: %r",
                                 self.syncDataFilepath, str(ex))
                    finally:
                        pickle_file.close()

            if self._sync_data is None:
                self._sync_data = {
                    # The version will be used to upgrade sync data if the
                    # format changes in the future.
                    'version': 1,
                    'local': {},
                    'remote': {},
                }
        return self._sync_data

    @sync_data.setter
    def sync_data(self, val):
        self._sync_data = val
        self._store_sync_data()

    def _store_sync_data(self):
        try:
            pickle_file = file(self.syncDataFilepath, "wb")
        except Exception, ex:
            log.warn("Could not write to publishing data file %r. Error: %r",
                     self.syncDataFilepath, str(ex))
        else:
            try:
                pickle.dump(self.sync_data, pickle_file)
            finally:
                pickle_file.close()

    def _updateSyncDataFromTransferOp(self, transferOp):
        # The download operation stores some private data that we need to get
        # our hands on, so we unwrap the transferOp and grab what we need.
        unwrappedOp = UnwrapObject(transferOp)
        localpaths = unwrappedOp._localpaths
        remotepaths = unwrappedOp._remotepaths
        removed_localpaths = unwrappedOp._removed_localpaths
        removed_remotepaths = unwrappedOp._removed_remotepaths

        localSyncInfos = [x and LocalSyncInfo(x) or None for x in unwrappedOp._local_stats]
        remoteSyncInfos = map(RemoteSyncInfo, unwrappedOp._rf_infos)

        # We have what we need - now update the stored sync data with this new
        # info.
        sync_data = self.sync_data

        local_sync_data = sync_data.get("local")
        new_data = dict(zip(localpaths, localSyncInfos))
        local_sync_data.update(new_data)
        for lpath in removed_localpaths:
            local_sync_data.pop(lpath, None)

        remote_sync_data = sync_data.get("remote")
        new_data = dict(zip(remotepaths, remoteSyncInfos))
        remote_sync_data.update(new_data)
        for rpath in removed_remotepaths:
            remote_sync_data.pop(rpath, None)

        # Save and store the data.
        self.sync_data = sync_data

    def updateDownloadSyncData(self, transferDownloadOp):
        self._updateSyncDataFromTransferOp(transferDownloadOp)

    def updateUploadSyncData(self, transferUploadOp):
        self._updateSyncDataFromTransferOp(transferUploadOp)
